#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import sys, os
pPath = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(pPath)
from PyQt5.QtWidgets import (QApplication, QWidget, QPushButton, QFileDialog, QLabel, QHBoxLayout, QGroupBox, QTextEdit,
                             QVBoxLayout, QTreeWidget, QTreeWidgetItem, QSplitter, QTableWidget, QTabWidget,
                             QTableWidgetItem, QInputDialog, QMessageBox, QFormLayout, QHeaderView, QAbstractItemView)
from PyQt5.QtGui import QIcon, QFont, QMovie
from PyQt5.QtCore import Qt, pyqtSignal
from util import (InputDialog, MachineLearning, ModelMetrics, PlotWidgets)
import qdarkstyle
import numpy as np
import pandas as pd
from itertools import combinations
import copy
import threading
import datetime
import joblib
import sip
import torch

class ILearnPlusAutoML(QWidget):
    # signal
    display_signal = pyqtSignal(list)
    display_curves_signal = pyqtSignal()
    append_msg_signal = pyqtSignal(str)
    close_signal = pyqtSignal(str)

    def __init__(self):
        super(ILearnPlusAutoML, self).__init__()
        # signal
        self.display_signal.connect(self.display_metrics)
        self.display_curves_signal.connect(self.display_curves)
        self.append_msg_signal.connect(self.append_message)

        self.MLData = None                                                     # Machine learning object
        self.algorithms_selected = set([])                                     # selected algorithms
        self.metrics = ModelMetrics.ModelMetrics()                             # data for display result
        self.current_data_index = 0
        self.boxplot_data = {}                                                 # dict
        self.pen_color = [
            (50, 116, 161),
            (225, 129, 44),
            (58, 146, 58),
            (192, 61, 62),
            (147, 114, 178),
            (132, 91, 83),
            (214, 132, 189),
            (127, 127, 127)
        ]
        self.lineStyle = {
            0: Qt.SolidLine,
            1: Qt.SolidLine,
            2: Qt.SolidLine,
            3: Qt.SolidLine,
            4: Qt.SolidLine,
            5: Qt.SolidLine,
            6: Qt.SolidLine,
            7: Qt.SolidLine,
            8: Qt.DashLine,
            9: Qt.DashLine,
            10: Qt.DashLine,
            11: Qt.DashLine,
            12: Qt.DashLine,
            13: Qt.DashLine,
            14: Qt.DashLine,
            15: Qt.DashLine,
        }
        self.ml_defatult_para = {
            'FOLD': 5,
            'cpu': 1,
            'auto': False,
            'n_trees': 100,
            'tree_range': (100, 1000, 100),
            'kernel': 'rbf',
            'penality': 1.0,
            'gamma': 'auto',
            'penalityRange': (1.0, 15.0),
            'gammaRange': (-10.0, 5.0),
            'layer': '32;32',
            'activation': 'relu',
            'optimizer': 'adam',
            'topKValue': 3,
            'boosting_type': 'gbdt',
            'num_leaves': 31,
            'max_depth': -1,
            'learning_rate': 0.01,
            'leaves_range': (20, 100, 10),
            'depth_range': (15, 55, 10),
            'rate_range': (0.01, 0.15, 0.02),
            'booster': 'gbtree',
            'n_estimator': 100,
            'colsample_bytree': 0.8,
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }                                          # default machine learning parameters
        self.para_rf = {
            'cpu': 1,
            'auto': False,
        }
        self.para_svm = {
            'auto': False,
        }
        self.para_lightgbm = {
            'max_depth': -1,
            'learning_rate': 0.01,
            'cpu': 1,
            'auto': False,
            'depth_range': (15, 55, 10),
            'rate_range': (0.01, 0.15, 0.02),
        }
        self.para_xgboost = {
            'max_depth': -1,
            'learning_rate': 0.01,
            'n_estimator': 100,
            'auto': False,
            'depth_range': (15, 55, 10),
            'rate_range': (0.01, 0.15, 0.02),
        }
        self.para_bagging = {
            'n_estimator': 100,
            'cpu': 1,
        }
        self.para_net_1 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_11 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_2 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_3 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_4 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_5 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.para_net_6 = {
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        }
        self.ml_combination_para = {
            'FOLD': 5,
            'cpu': 1,
            'auto': True,
            'n_trees': 100,
            'tree_range': (100, 1000, 100),
            'kernel': 'rbf',
            'penality': 1.0,
            'gamma': 'auto',
            'penalityRange': (1.0, 15.0),
            'gammaRange': (-10.0, 5.0),
            'layer': '32;32',
            'activation': 'relu',
            'optimizer': 'adam',
            'topKValue': 3,
            'boosting_type': 'gbdt',
            'num_leaves': 31,
            'max_depth': -1,
            'learning_rate': 0.01,
            'leaves_range': (20, 100, 10),
            'depth_range': (15, 55, 10),
            'rate_range': (0.01, 0.15, 0.02),
            'booster': 'gbtree',
            'n_estimator': 100,
            'colsample_bytree': 0.8,
            'input_channel': 1,
            'input_length': 100,
            'output_channel': 64,
            'padding': 2,
            'kernel_size': 5,
            'dropout': 0.5,
            'epochs': 1000,
            'early_stopping': 100,
            'batch_size': 64,
            'fc_size': 64,
            'rnn_hidden_size': 32,
            'rnn_hidden_layers': 1,
            'rnn_bidirection': False,
            'mlp_input_dim': None,
        } 

        # for combination of models
        self.MLData_cb = None     
        self.bestPerformance = 0
        self.bestCombinations = ()
        self.bestMetrics = None
        self.bestAUC = None
        self.bestPRC = None
        self.bestModels = None
        self.bestTrainingScore = None

        # status bar
        self.gif = QMovie(os.path.join(pPath, 'images', 'progress_bar.gif'))
        self.initUI()

    def initUI(self):
        self.setWindowTitle('iLearnPlus AutoML')
        self.resize(800, 600)
        self.setWindowState(Qt.WindowMaximized)
        self.setWindowIcon(QIcon(os.path.join(pPath, 'images', 'logo.ico')))
        self.setup_UI()

    def setup_UI(self):
        # file
        topGroupBox = QGroupBox('Load data', self)
        topGroupBox.setFont(QFont('Arial', 10))
        # topGroupBox.setMinimumHeight(100)
        topGroupBoxLayout = QFormLayout()
        trainFileButton = QPushButton('Open')
        trainFileButton.clicked.connect(lambda:self.data_from_file_ml('Training'))
        # testFileButton = QPushButton('Open')
        # testFileButton.clicked.connect(lambda:self.data_from_file_ml('Testing'))
        topGroupBoxLayout.addRow('Open training file:', trainFileButton)
        # topGroupBoxLayout.addRow('Open testing file:', testFileButton)
        topGroupBox.setLayout(topGroupBoxLayout)

        # tree
        treeGroupBox = QGroupBox('Machine learning algorithms', self)
        treeGroupBox.setFont(QFont('Arial', 10))
        treeLayout = QHBoxLayout()
        self.ml_treeWidget = QTreeWidget()
        self.ml_treeWidget.setColumnCount(2)
        self.ml_treeWidget.setMinimumWidth(300)
        self.ml_treeWidget.setColumnWidth(0, 150)
        self.ml_treeWidget.setFont(QFont('Arial', 8))
        self.ml_treeWidget.setHeaderLabels(['Methods', 'Definition'])
        self.machineLearningAlgorighms = QTreeWidgetItem(self.ml_treeWidget)
        self.machineLearningAlgorighms.setExpanded(True)  # set node expanded
        self.machineLearningAlgorighms.setText(0, 'Machine learning algorithms')
        rf = QTreeWidgetItem(self.machineLearningAlgorighms)
        rf.setText(0, 'RF')
        rf.setText(1, 'Random Forest')
        rf.setCheckState(0, Qt.Unchecked)
        dtree = QTreeWidgetItem(self.machineLearningAlgorighms)
        dtree.setText(0, 'DecisionTree')
        dtree.setText(1, 'Decision Tree')
        dtree.setCheckState(0, Qt.Unchecked)
        lightgbm = QTreeWidgetItem(self.machineLearningAlgorighms)
        lightgbm.setText(0, 'LightGBM')
        lightgbm.setText(1, 'LightGBM')
        lightgbm.setCheckState(0, Qt.Unchecked)
        svm = QTreeWidgetItem(self.machineLearningAlgorighms)
        svm.setText(0, 'SVM')
        svm.setText(1, 'Support Verctor Machine')
        svm.setCheckState(0, Qt.Unchecked)
        mlp = QTreeWidgetItem(self.machineLearningAlgorighms)
        mlp.setText(0, 'MLP')
        mlp.setText(1, 'Multi-layer Perceptron')
        mlp.setCheckState(0, Qt.Unchecked)
        xgboost = QTreeWidgetItem(self.machineLearningAlgorighms)
        xgboost.setText(0, 'XGBoost')
        xgboost.setText(1, 'XGBoost')
        xgboost.setCheckState(0, Qt.Unchecked)
        knn = QTreeWidgetItem(self.machineLearningAlgorighms)
        knn.setText(0, 'KNN')
        knn.setText(1, 'K-Nearest Neighbour')
        knn.setCheckState(0, Qt.Unchecked)
        lr = QTreeWidgetItem(self.machineLearningAlgorighms)
        lr.setText(0, 'LR')
        lr.setText(1, 'Logistic Regression')
        lr.setCheckState(0, Qt.Unchecked)
        lda = QTreeWidgetItem(self.machineLearningAlgorighms)
        lda.setText(0, 'LDA')
        lda.setText(1, 'Linear Discriminant Analysis')
        lda.setCheckState(0, Qt.Unchecked)
        qda = QTreeWidgetItem(self.machineLearningAlgorighms)
        qda.setText(0, 'QDA')
        qda.setText(1, 'Quadratic Discriminant Analysis')
        qda.setCheckState(0, Qt.Unchecked)
        sgd = QTreeWidgetItem(self.machineLearningAlgorighms)
        sgd.setText(0, 'SGD')
        sgd.setText(1, 'Stochastic Gradient Descent')
        sgd.setCheckState(0, Qt.Unchecked)
        bayes = QTreeWidgetItem(self.machineLearningAlgorighms)
        bayes.setText(0, 'NaiveBayes')
        bayes.setText(1, 'NaiveBayes')
        bayes.setCheckState(0, Qt.Unchecked)
        bagging = QTreeWidgetItem(self.machineLearningAlgorighms)
        bagging.setText(0, 'Bagging')
        bagging.setText(1, 'Bagging')
        bagging.setCheckState(0, Qt.Unchecked)
        adaboost = QTreeWidgetItem(self.machineLearningAlgorighms)
        adaboost.setText(0, 'AdaBoost')
        adaboost.setText(1, 'AdaBoost')
        adaboost.setCheckState(0, Qt.Unchecked)
        gbdt = QTreeWidgetItem(self.machineLearningAlgorighms)
        gbdt.setText(0, 'GBDT')
        gbdt.setText(1, 'Gradient Tree Boosting')
        gbdt.setCheckState(0, Qt.Unchecked)
        # deep learning algorighms
        net1 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net1.setText(0, 'Net_1_CNN')
        net1.setText(1, 'Convolutional Neural Network')
        net1.setCheckState(0, Qt.Unchecked)
        net11 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net11.setText(0, 'Net_1_CNN_binary')
        net11.setText(1, 'Convolutional Neural Network')
        net11.setCheckState(0, Qt.Unchecked)
        net2 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net2.setText(0, 'Net_2_RNN')
        net2.setText(1, 'Recurrent Neural Network')
        net2.setCheckState(0, Qt.Unchecked)
        net3 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net3.setText(0, 'Net_3_BRNN')
        net3.setText(1, 'Bidirectional Recurrent Neural Network')
        net3.setCheckState(0, Qt.Unchecked)
        net4 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net4.setText(0, 'Net_4_ABCNN')
        net4.setText(1, 'Attention Based Convolutional Neural Network')
        net4.setCheckState(0, Qt.Unchecked)
        net5 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net5.setText(0, 'Net_5_ResNet')
        net5.setText(1, 'Deep Residual Network')
        net5.setCheckState(0, Qt.Unchecked)
        net6 = QTreeWidgetItem(self.machineLearningAlgorighms)
        net6.setText(0, 'Net_6_AE')
        net6.setText(1, 'AutoEncoder')
        net6.setCheckState(0, Qt.Unchecked)
        treeLayout.addWidget(self.ml_treeWidget)
        treeGroupBox.setLayout(treeLayout)
        self.ml_treeWidget.clicked.connect(self.ml_tree_clicked)
        self.ml_treeWidget.itemChanged.connect(self.ml_tree_checkState)

        ## parameter
        paraGroupBox = QGroupBox('Set K-fold Cross-Validation', self)
        paraGroupBox.setFont(QFont('Arial', 10))
        paraLayout = QFormLayout(paraGroupBox)
        self.fold_lineEdit = InputDialog.MyLineEdit('5')
        self.fold_lineEdit.setFont(QFont('Arial', 8))
        self.fold_lineEdit.clicked.connect(self.setFold)
        paraLayout.addRow('Cross-Validation:', self.fold_lineEdit)

        # operation
        startGroupBox = QGroupBox('Operator', self)
        startGroupBox.setFont(QFont('Arial', 10))
        startLayout = QHBoxLayout(startGroupBox)
        self.start_button = QPushButton('Start')
        self.start_button.clicked.connect(self.run_ml)
        self.start_button.setFont(QFont('Arial', 10))
        startLayout.addWidget(self.start_button)

        ### layout
        left_vertical_layout = QVBoxLayout()
        left_vertical_layout.addWidget(topGroupBox)
        left_vertical_layout.addWidget(treeGroupBox)
        left_vertical_layout.addWidget(paraGroupBox)
        left_vertical_layout.addWidget(startGroupBox)

        #### widget
        leftWidget = QWidget()
        leftWidget.setLayout(left_vertical_layout)

        resultTabWidget = QWidget()
        resultTabLayout = QVBoxLayout(resultTabWidget)
        resultTabControlLayout = QHBoxLayout()
        self.resultSaveBtn = QPushButton('Save metrics')
        self.resultSaveBtn.clicked.connect(self.save_result)
        self.modelSaveBtn = QPushButton('Save model')
        self.modelSaveBtn.clicked.connect(self.save_model)
        self.displayCorrBtn = QPushButton(' Display Correlation ')
        self.displayCorrBtn.clicked.connect(self.display_correlation_heatmap)
        self.combineModelBtn = QPushButton(' Combine models ')
        self.combineModelBtn.clicked.connect(self.combineModels)

        resultTabControlLayout.addStretch(1)
        resultTabControlLayout.addWidget(self.resultSaveBtn)
        resultTabControlLayout.addWidget(self.modelSaveBtn)
        resultTabControlLayout.addWidget(self.displayCorrBtn)
        resultTabControlLayout.addWidget(self.combineModelBtn)
        resultTabControlLayout.addStretch(1)
        self.metricsTableWidget = QTableWidget()
        self.metricsTableWidget.setFont(QFont('Arial', 8))
        self.metricsTableWidget.setColumnCount(11)
        self.metricsTableWidget.setHorizontalHeaderLabels(['Id', 'StartTime', 'EndTime', 'Sn (%)', 'Sp (%)', 'Pre (%)', 'Acc (%)', 'MCC', 'F1', 'AUROC', 'AUPRC'])
        self.metricsTableWidget.verticalHeader().setHidden(True)
        self.metricsTableWidget.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.metricsTableWidget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.metricsTableWidget.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.metricsTableWidget.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.metricsTableWidget.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.metricsTableWidget.resizeRowsToContents()
        resultTabLayout.addWidget(self.metricsTableWidget)
        resultTabLayout.addLayout(resultTabControlLayout)

        """ boxplot using matplotlib """
        boxplotWidget = QWidget()
        self.boxplotLayout = QVBoxLayout(boxplotWidget)
        self.boxplotGraph = PlotWidgets.BoxplotWidget()
        self.boxplotLayout.addWidget(self.boxplotGraph)

        """ ROC curve using matplotlib """
        rocWidget = QWidget()
        self.rocLayout = QVBoxLayout(rocWidget)
        self.rocCurveGraph = PlotWidgets.CurvesWidget()
        self.rocLayout.addWidget(self.rocCurveGraph)

        logWidget = QWidget()
        logLayout = QHBoxLayout(logWidget)
        self.logTextEdit = QTextEdit()
        self.logTextEdit.setFont(QFont('Arial', 8))
        logLayout.addWidget(self.logTextEdit)

        plotTabWidget = QTabWidget()
        plotTabWidget.addTab(resultTabWidget, '  Result  ')
        plotTabWidget.addTab(boxplotWidget, '  Box plot  ')
        plotTabWidget.addTab(rocWidget, '  ROC and PRC curve  ')
        plotTabWidget.addTab(logWidget, '  Log  ')

        ##### splitter
        splitter_1 = QSplitter(Qt.Horizontal)
        splitter_1.addWidget(leftWidget)
        splitter_1.addWidget(plotTabWidget)
        splitter_1.setSizes([100, 1200])

        ###### vertical layout
        vLayout = QVBoxLayout()

        ## status bar
        statusGroupBox = QGroupBox('Status', self)
        statusGroupBox.setFont(QFont('Arial', 10))
        statusLayout = QHBoxLayout(statusGroupBox)
        self.status_label = QLabel('Welcome to iLearnPlus Analysis')
        self.progress_bar = QLabel()
        self.progress_bar.setMaximumWidth(230)
        statusLayout.addWidget(self.status_label)
        statusLayout.addWidget(self.progress_bar)

        splitter_2 = QSplitter(Qt.Vertical)
        splitter_2.addWidget(splitter_1)
        splitter_2.addWidget(statusGroupBox)
        splitter_2.setSizes([1000, 100])
        vLayout.addWidget(splitter_2)
        self.setLayout(vLayout)

    def data_from_file_ml(self, target='Training'):
        file_name, ok = QFileDialog.getOpenFileName(self, 'Open', os.path.join(pPath, 'data'), 'CSV Files (*.csv);;TSV Files (*.tsv);;SVM Files(*.svm);;Weka Files (*.arff)')
        if ok:
            if self.MLData is None:
                self.MLData = MachineLearning.ILearnMachineLearning(self.ml_defatult_para)
            ok1 = self.MLData.load_data(file_name, target)
            if ok1:
                if target == 'Training':
                    shape = self.MLData.training_dataframe.values.shape
                else:
                    shape = self.MLData.testing_dataframe.values.shape
                self.logTextEdit.append('%s\tLoad %s file; Datashape: %s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), target, shape))
            else:
                self.logTextEdit.append('%s\tLoad %s file failed.' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), target))

    def ml_tree_clicked(self, index):
        item = self.ml_treeWidget.currentItem()
        if item.text(0) in ['RF']:
            num, range, cpu, auto, ok = InputDialog.QRandomForestInput.getValues()
            if ok:
                self.ml_defatult_para['n_trees'] = num
                self.ml_defatult_para['tree_range'] = range
                # self.ml_defatult_para['auto'] = auto
                # self.ml_defatult_para['cpu'] = cpu
                self.para_rf['auto'] = auto
                self.para_rf['cpu'] = cpu
                if auto:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[Auto-optimization\ttree_range=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), str(range)))
                else:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[tree number=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), num))
        elif item.text(0) in ['SVM']:
            kernel, penality, gamma, auto, penalityRange, gammaRange, ok = InputDialog.QSupportVectorMachineInput.getValues()
            if ok:
                self.ml_defatult_para['kernel'] = kernel
                self.ml_defatult_para['penality'] = penality
                self.ml_defatult_para['gamma'] = gamma
                # self.ml_defatult_para['auto'] = auto
                self.para_svm['auto'] = auto
                self.ml_defatult_para['penalityRange'] = penalityRange
                self.ml_defatult_para['gammaRange'] = gammaRange
                if auto:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[kernel=%s\tAuto-optimization\tpenality range=%s\tgamma range=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), kernel, str(penalityRange), str(gammaRange)))
                else:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[kernel=%s\tpenality=%s\tgamma=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), kernel, penality, gamma))
        elif item.text(0) in ['MLP']:
            layer, epochs, activation, optimizer, ok = InputDialog.QMultiLayerPerceptronInput.getValues()
            if ok:
                self.ml_defatult_para['layer'] = layer
                self.ml_defatult_para['epochs'] = epochs
                self.ml_defatult_para['activation'] = activation
                self.ml_defatult_para['optimizer'] = optimizer
                self.logTextEdit.append('%s\tAlgorithm: %s\t[layer=%s\tepochs=%s\tactivation function=%s\toptimizer=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), layer, epochs, activation, optimizer))
        elif item.text(0) in ['LR', 'SGD', 'DecisionTree', 'NaiveBayes', 'AdaBoost', 'GBDT', 'LDA', 'QDA']:
            self.logTextEdit.append('%s\tAlgorithm: %s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0)))
        elif item.text(0) in ['KNN']:
            topKValue, ok = InputDialog.QKNeighborsInput.getValues()
            if ok:
                self.ml_defatult_para['topKValue'] = topKValue
                self.logTextEdit.append('%s\tAlgorithm: %s\t[top K value=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), topKValue))
        elif item.text(0) in ['LightGBM']:
            type, leaves, depth, rate, leavesRange, depthRange, rateRange, threads, auto, ok = InputDialog.QLightGBMInput.getValues()
            if ok:
                self.ml_defatult_para['boosting_type'] = type
                self.ml_defatult_para['num_leaves'] = leaves
                # self.ml_defatult_para['max_depth'] = depth
                # self.ml_defatult_para['learning_rate'] = rate
                # self.ml_defatult_para['auto'] = auto
                self.ml_defatult_para['leaves_range'] = leavesRange
                # self.ml_defatult_para['depth_range'] = depthRange
                # self.ml_defatult_para['rate_range'] = rateRange
                # self.ml_defatult_para['cpu'] = threads
                self.para_lightgbm['max_depth'] = depth
                self.para_lightgbm['learning_rate'] = rate
                self.para_lightgbm['auto'] = auto
                self.para_lightgbm['depth_range'] = depthRange
                self.para_lightgbm['rate_range'] = rateRange
                self.para_lightgbm['cpu'] = threads
                if auto:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[Auto optimization\tLeaves range=%s\tDepth range=%s\tLearning rate range=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), leavesRange, depthRange, rateRange))
                else:
                    self.logTextEdit.append('%s\tAlgorithm: %s\t[boosting type=%s\tnumber of leaves=%s\tmax depth=%s\tlearning rate=%s\tnumber of threads=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), type, leaves, depth, rate, threads))
        elif item.text(0) in ['XGBoost']:
            booster, maxdepth, rate, estimator, colsample, depthRange, rateRange, threads, auto, ok = InputDialog.QXGBoostInput.getValues()
            self.ml_defatult_para['booster'] = booster
            # self.ml_defatult_para['max_depth'] = maxdepth
            # self.ml_defatult_para['learning_rate'] = rate
            # self.ml_defatult_para['n_estimator'] = estimator
            self.ml_defatult_para['colsample_bytree'] = colsample
            # self.ml_defatult_para['depth_range'] = depthRange
            # self.ml_defatult_para['rate_range'] = rateRange
            # self.ml_defatult_para['cpu'] = threads
            # self.ml_defatult_para['auto'] = auto
            self.para_xgboost['max_depth'] = maxdepth
            self.para_xgboost['learning_rate'] = rate
            self.para_xgboost['n_estimator'] = estimator
            self.para_xgboost['depth_range'] = depthRange
            self.para_xgboost['rate_range'] = rateRange
            self.para_xgboost['cpu'] = threads
            self.para_xgboost['auto'] = auto
            if auto:
                self.logTextEdit.append('%s\tAlgorithm: %s\t[Auto optimization\tDepth range=%s\tLearning rate range=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), depthRange, rateRange))
            else:
                self.logTextEdit.append('%s\tAlgorithm: %s\t[Booster=%s\tMax depth=%s\tLearning rate=%s\tn_estimator=%s\tcolsample_bytree=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), booster, maxdepth, rate, estimator, colsample))
        elif item.text(0) in ['Bagging']:
            n_estimators, threads, ok = InputDialog.QBaggingInput.getValues()
            if ok:
                self.para_bagging['n_estimator'] = n_estimators
                self.para_bagging['cpu'] = threads
                self.logTextEdit.append('%s\tAlgorithm: %s\t[n_estimators=%s\tThreads=%s]' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), n_estimators, threads))
        elif item.text(0) in ['Net_1_CNN']:
            if not self.MLData is None:
                input_channel, input_length, output_channel, padding, kernel_size, dropout, learning_rate, epochs, early_stopping, batch_size, fc_size, ok = InputDialog.QNetInput_1.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_1['input_channel'] = input_channel
                    self.para_net_1['input_length'] = input_length
                    self.para_net_1['output_channel'] = output_channel
                    self.para_net_1['padding'] = padding
                    self.para_net_1['kernel_size'] = kernel_size
                    self.para_net_1['dropout'] = dropout
                    self.para_net_1['learning_rate'] = learning_rate
                    self.para_net_1['epochs'] = epochs
                    self.para_net_1['early_stopping'] = early_stopping
                    self.para_net_1['batch_size'] = batch_size
                    self.para_net_1['fc_size'] = fc_size
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input channel=%s; Input_length=%s; Output_channel=%s; Padding=%s; Kernel_size=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, output_channel, padding, kernel_size, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_1_CNN_binary']:
            if not self.MLData is None:
                input_channel, input_length, output_channel, padding, kernel_size, dropout, learning_rate, epochs, early_stopping, batch_size, fc_size, ok = InputDialog.QNetInput_1.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_11['input_channel'] = input_channel
                    self.para_net_11['input_length'] = input_length
                    self.para_net_11['output_channel'] = output_channel
                    self.para_net_11['padding'] = padding
                    self.para_net_11['kernel_size'] = kernel_size
                    self.para_net_11['dropout'] = dropout
                    self.para_net_11['learning_rate'] = learning_rate
                    self.para_net_11['epochs'] = epochs
                    self.para_net_11['early_stopping'] = early_stopping
                    self.para_net_11['batch_size'] = batch_size
                    self.para_net_11['fc_size'] = fc_size
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input channel=%s; Input_length=%s; Output_channel=%s; Padding=%s; Kernel_size=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, output_channel, padding, kernel_size, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_2_RNN']:
            if not self.MLData is None:
                input_channel, input_length, hidden_size, num_layers, dropout, learning_rate, epochs, early_stopping, batch_size, fc_size, ok = InputDialog.QNetInput_2.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_2['input_channel'] = input_channel
                    self.para_net_2['input_length'] = input_length
                    self.para_net_2['rnn_hidden_size'] = hidden_size
                    self.para_net_2['rnn_hidden_layers'] = num_layers
                    self.para_net_2['rnn_bidirection'] = False
                    self.para_net_2['dropout'] = dropout
                    self.para_net_2['learning_rate'] = learning_rate
                    self.para_net_2['epochs'] = epochs
                    self.para_net_2['early_stopping'] = early_stopping
                    self.para_net_2['batch_size'] = batch_size
                    self.para_net_2['rnn_bidirectional'] = False
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input size=%s; Input_length=%s; Hidden_size=%s; Num_hidden_layers=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, hidden_size, num_layers, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_3_BRNN']:
            if not self.MLData is None:
                input_channel, input_length, hidden_size, num_layers, dropout, learning_rate, epochs, early_stopping, batch_size, fc_size, ok = InputDialog.QNetInput_2.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_3['input_channel'] = input_channel
                    self.para_net_3['input_length'] = input_length
                    self.para_net_3['rnn_hidden_size'] = hidden_size
                    self.para_net_3['rnn_hidden_layers'] = num_layers
                    self.para_net_3['rnn_bidirection'] = False
                    self.para_net_3['dropout'] = dropout
                    self.para_net_3['learning_rate'] = learning_rate
                    self.para_net_3['epochs'] = epochs
                    self.para_net_3['early_stopping'] = early_stopping
                    self.para_net_3['batch_size'] = batch_size
                    self.para_net_3['rnn_bidirectional'] = True
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input size=%s; Input_length=%s; Hidden_size=%s; Num_hidden_layers=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, hidden_size, num_layers, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_4_ABCNN']:
            if not self.MLData is None:
                input_channel, input_length, dropout, learning_rate, epochs, early_stopping, batch_size, ok = InputDialog.QNetInput_4.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_4['input_channel'] = input_channel
                    self.para_net_4['input_length'] = input_length
                    self.para_net_4['dropout'] = dropout
                    self.para_net_4['learning_rate'] = learning_rate
                    self.para_net_4['epochs'] = epochs
                    self.para_net_4['early_stopping'] = early_stopping
                    self.para_net_4['batch_size'] = batch_size
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input size=%s; Input_length=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_5_ResNet']:
            if not self.MLData is None:
                input_channel, input_length, learning_rate, epochs, early_stopping, batch_size, ok = InputDialog.QNetInput_5.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_5['input_channel'] = input_channel
                    self.para_net_5['input_length'] = input_length
                    self.para_net_5['learning_rate'] = learning_rate
                    self.para_net_5['epochs'] = epochs
                    self.para_net_5['early_stopping'] = early_stopping
                    self.para_net_5['batch_size'] = batch_size
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input size=%s; Input_length=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s'  %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_channel, input_length, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif item.text(0) in ['Net_6_AE']:
            if not self.MLData is None:
                input_dim, dropout, learning_rate, epochs, early_stopping, batch_size, ok = InputDialog.QNetInput_6.getValues(self.MLData.training_dataframe.values.shape[1])
                if ok:
                    self.para_net_6['mlp_input_dim'] = input_dim
                    self.para_net_6['dropout'] = dropout
                    self.para_net_6['learning_rate'] = learning_rate
                    self.para_net_6['epochs'] = epochs
                    self.para_net_6['early_stopping'] = early_stopping
                    self.para_net_6['batch_size'] = batch_size
                    self.logTextEdit.append('%s\tAlgorithm: %s; Input dimension=%s; Dropout=%s; lr=%s; Epochs=%s; Early_stopping=%s; Batch_size=%s' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), item.text(0), input_dim, dropout, learning_rate, epochs, early_stopping, batch_size))
            else:
                QMessageBox.warning(self, 'Warning', 'Please input training data at first!', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def ml_tree_checkState(self, item, column):
        if item and item.text(0) not in ['Machine learning algorithms']:
            self.ml_treeWidget.setCurrentItem(item)
            if item.checkState(column) == Qt.Checked:
                self.algorithms_selected.add(item.text(0))
            if item.checkState(column) == Qt.Unchecked:
                self.algorithms_selected.discard(item.text(0))

    def setFold(self):
        fold, ok = QInputDialog.getInt(self, 'Fold number', 'Setting K-fold cross-validation', 5, 2, 100, 1)
        if ok:
            self.fold_lineEdit.setText(str(fold))
            self.ml_defatult_para['FOLD'] = fold
            self.logTextEdit.append('%s\tSet %s-fold Cross-validation' %(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), fold))

    def panel_clear(self):
        # metricsTableWidget init
        self.metricsTableWidget.clear()
        self.metricsTableWidget.setColumnCount(11)
        self.metricsTableWidget.setRowCount(0)
        self.current_data_index = 0
        self.metricsTableWidget.setHorizontalHeaderLabels(['Id', 'StartTime', 'EndTime', 'Sn (%)', 'Sp (%)', 'Pre (%)', 'Acc (%)', 'MCC', 'F1', 'AUROC', 'AUPRC'])
        self.descriptor = None
        self.fasta_file = None
        #
        self.metrics = ModelMetrics.ModelMetrics()
        self.boxplot_data = {}

    def restart_init(self):
        # metricsTableWidget init
        self.metricsTableWidget.clear()
        self.metricsTableWidget.setColumnCount(11)
        self.metricsTableWidget.setRowCount(0)
        self.current_data_index = 0
        self.metricsTableWidget.setHorizontalHeaderLabels(['Id', 'StartTime', 'EndTime', 'Sn (%)', 'Sp (%)', 'Pre (%)', 'Acc (%)', 'MCC', 'F1', 'AUROC', 'AUPRC'])
        self.metrics = ModelMetrics.ModelMetrics()
        self.boxplot_data = {}
        self.combineModelBtn.setDisabled(False)

    def run_ml(self):
        self.setDisabled(True)
        self.restart_init()
        if self.MLData is None:
            QMessageBox.critical(self, 'Error', 'Please load training file.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif len(self.algorithms_selected) == 0:
            QMessageBox.critical(self, 'Error', 'Please select at least one machine learning algorithm.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        elif self.ml_defatult_para['FOLD'] <= 1:
            QMessageBox.critical(self, 'Error', 'FOLD need > 1', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        else:
            self.logTextEdit.append('%s\tJobs start ...' %datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
            self.progress_bar.setMovie(self.gif)
            self.gif.start()
            t = threading.Thread(target=self.train_model)
            t.start()

    def train_model(self):
        for algo in self.algorithms_selected:
            self.status_label.setText('Start training %s model  ...' % algo)
            start_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            if algo == 'RF':
                for key in self.para_rf:
                    self.ml_defatult_para[key] = self.para_rf[key]                    
                ok = self.MLData.RandomForest()
            elif algo == 'SVM':
                for key in self.para_svm:
                    self.ml_defatult_para[key] = self.para_svm[key]
                ok = self.MLData.SupportVectorMachine()
            elif algo == 'MLP':
                ok = self.MLData.MultiLayerPerceptron()
            elif algo == 'LR':
                ok = self.MLData.LogisticRegressionClassifier()
            elif algo == 'LDA':
                ok = self.MLData.LDAClassifier()
            elif algo == 'QDA':
                ok = self.MLData.QDAClassifier()
            elif algo == 'KNN':
                ok = self.MLData.KNeighbors()
            elif algo == 'LightGBM':
                for key in self.para_lightgbm:
                    self.ml_defatult_para[key] = self.para_lightgbm[key]
                ok = self.MLData.LightGBMClassifier()
            elif algo == 'XGBoost':
                for key in self.para_xgboost:
                    self.ml_defatult_para[key] = self.para_xgboost[key]
                ok = self.MLData.XGBoostClassifier()
            elif algo == 'SGD':
                ok = self.MLData.StochasticGradientDescentClassifier()
            elif algo == 'DecisionTree':
                ok = self.MLData.DecisionTree()
            elif algo == 'NaiveBayes':
                ok = self.MLData.GaussianNBClassifier()
            elif algo == 'AdaBoost':
                ok = self.MLData.AdaBoost()
            elif algo == 'Bagging':
                for key in self.para_bagging:
                    self.ml_defatult_para[key] = self.para_bagging[key]
                ok = self.MLData.Bagging()
            elif algo == 'GBDT':
                ok = self.MLData.GBDTClassifier()
            elif algo == 'Net_1_CNN':
                for key in self.para_net_1:
                    self.ml_defatult_para[key] = self.para_net_1[key]
                ok = self.MLData.run_networks(1)
            elif algo == 'Net_1_CNN_binary':
                for key in self.para_net_11:
                    self.ml_defatult_para[key] = self.para_net_11[key]
                ok = self.MLData.run_networks(11)
            elif algo == 'Net_2_RNN':
                for key in self.para_net_2:
                    self.ml_defatult_para[key] = self.para_net_2[key]
                ok = self.MLData.run_networks(2)
            elif algo == 'Net_3_BRNN':
                for key in self.para_net_3:
                    self.ml_defatult_para[key] = self.para_net_3[key]
                ok = self.MLData.run_networks(3)
            elif algo == 'Net_4_ABCNN':
                for key in self.para_net_4:
                    self.ml_defatult_para[key] = self.para_net_4[key]
                ok = self.MLData.run_networks(4)
            elif algo == 'Net_5_ResNet':
                for key in self.para_net_5:
                    self.ml_defatult_para[key] = self.para_net_5[key]
                ok = self.MLData.run_networks(5)
            elif algo == 'Net_6_AE':
                for key in self.para_net_6:
                    self.ml_defatult_para[key] = self.para_net_6[key]
                ok = self.MLData.run_networks(6)

            self.MLData.calculate_boxplot_data()
            end_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            if ok:
                self.status_label.setText('%s model training complete.' %algo)
                data_list = ['%s_model' %algo, start_time, end_time] + list(self.MLData.metrics.values[-1])
                item = pd.DataFrame([data_list], columns=['Id', 'StartTime', 'EndTime', 'Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'])
                self.metrics.insert_data(item, '%s_model' %algo, self.MLData.meanAucData, self.MLData.meanPrcData, self.MLData.training_score, self.MLData.best_model)
                self.boxplot_data[algo] = self.MLData.metrics.iloc[0:self.ml_defatult_para['FOLD'], :]
                self.display_signal.emit(data_list)
            else:
                self.status_label.setText('%s model training failed.' %algo)
        self.display_curves_signal.emit()

    def display_metrics(self, data_list):
        index = self.current_data_index
        self.metricsTableWidget.insertRow(index)
        self.current_data_index += 1
        for i in range(len(data_list)):
            self.metricsTableWidget.setItem(index, i, QTableWidgetItem(str(data_list[i])))
        self.metricsTableWidget.resizeRowsToContents()

    def display_curves(self):
        self.status_label.setText('Plotting ... ')
        self.rocLayout.removeWidget(self.rocCurveGraph)
        sip.delete(self.rocCurveGraph)
        self.rocCurveGraph = PlotWidgets.CurvesWidget()
        self.rocLayout.addWidget(self.rocCurveGraph)
        try:
            if self.MLData.task == 'binary':
                self.rocCurveGraph.init_roc_data(0, 'ROC curve', self.metrics.aucData)
            if self.MLData.task == 'binary':
                self.rocCurveGraph.init_prc_data(1, 'PRC curve', self.metrics.prcData)
        except Exception as e:
            QMessageBox.warning(self, 'Warning', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

        """ boxplot using matplotlib """
        self.boxplotLayout.removeWidget(self.boxplotGraph)
        sip.delete(self.boxplotGraph)
        self.boxplotGraph = PlotWidgets.BoxplotWidget()
        self.boxplotLayout.addWidget(self.boxplotGraph)
        try:
            if not self.MLData is None and self.MLData.task == 'binary':
                self.boxplotGraph.init_data(['Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'], self.boxplot_data)
            if not self.MLData is None and self.MLData.task == 'muti-task':
                self.boxplotGraph.init_data(['Acc'], self.boxplot_data)
        except Exception as e:
            QMessageBox.warning(self, 'Warning', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

        # init data for p_value calculation (bootstrap test)
        self.rocCurveGraph.init_prediction_scores(self.MLData.task, self.metrics.prediction_scores)

        self.gif.stop()
        self.progress_bar.clear()
        self.start_button.setDisabled(False)
        self.setDisabled(False)
        self.status_label.setText('Operation complete.')

    def save_result(self):
        try:
            saved_file, ok = QFileDialog.getSaveFileName(self, 'Save', './data', 'TSV Files (*.tsv)')
            if ok:
                self.metrics.metrics.to_csv(saved_file, sep='\t', header=True, index=False)
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def save_model(self):
        try:
            model_list = self.metrics.metrics.loc[:, 'Id'].values.tolist()
            if len(model_list) > 0:
                model_id, ok = InputDialog.QSelectModel.getValues(model_list)
                if ok:
                    if model_id in self.metrics.models:
                        save_directory = QFileDialog.getExistingDirectory(self, 'Save', './data')
                        if not os.path.exists(save_directory):
                            pass
                        else:
                            for i, model in enumerate(self.metrics.models[model_id]):
                                model_name = '%s/%s_%s.pkl' %(save_directory, model_id, i+1)
                                if 'predict_proba' in dir(model):
                                    joblib.dump(model, model_name)
                                else:
                                    torch.save(model, model_name)
                            QMessageBox.information(self, 'Model saved',
                                                    'The models have been saved to directory %s' % save_directory,
                                                    QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
                    else:
                        QMessageBox.critical(self, 'Error', 'An error has been encountered in saving the model.',
                                            QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
            else:
                QMessageBox.critical(self, 'Error', 'No model can be saved.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def combineModels(self):
        model_list = self.metrics.metrics.loc[:, 'Id'].values.tolist()
        if len(model_list) >= 2:
            combineAlgorithm, ok = InputDialog.QSCombineModelDialog.getValues(model_list)
            combinations_array = []
            if ok:
                for model_num in range(2, len(model_list) + 1):
                    combinations_array += list(combinations(model_list, model_num))
                self.logTextEdit.append('%s\tStarting train combined models ...' %datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                self.progress_bar.setMovie(self.gif)
                self.gif.start()
                t = threading.Thread(target=lambda : self.calculateCombinations(combineAlgorithm, combinations_array))
                t.start()
        else:
            QMessageBox.critical(self, 'Error', 'At least two models needed.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def calculateCombinations(self, combineAlgorithm, combinations_array):
        self.setDisabled(True)
        start_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        self.ml_combination_para['FOLD'] = self.ml_defatult_para['FOLD']        
        for c in combinations_array:
            self.status_label.setText('Training combined model for combination %s with %s method.' %(str(c), combineAlgorithm))
            msg = '%s\tTraining model for combinations %s with %s method ...' %(start_time, str(c), combineAlgorithm)
            self.append_msg_signal.emit(msg)
            # if self.metrics.classification_task == 'binary':
            #     ensemble_data_X = np.concatenate(tuple(self.metrics.prediction_scores[m].values[:, 3].reshape((-1, 1)) for m in c), axis=1)
            # else:
            #     ensemble_data_X = np.concatenate(tuple(self.metrics.prediction_scores[m].values[:, 2:] for m in c), axis=1)
            ensemble_data_X = np.concatenate(tuple(self.metrics.prediction_scores[m].values[:, 2:] for m in c), axis=1)
            ensemble_data_y = self.metrics.prediction_scores[c[0]].values[:, 1]            
            df = pd.DataFrame(ensemble_data_X.astype(float))
            label = ensemble_data_y.astype(int)
            self.MLData_cb = MachineLearning.ILearnMachineLearning(self.ml_combination_para)
            self.MLData_cb.import_training_data(df, label)
            
            if combineAlgorithm == 'RF':
                ok = self.MLData_cb.RandomForest()
            elif combineAlgorithm == 'SVM':
                ok = self.MLData_cb.SupportVectorMachine()
            elif combineAlgorithm == 'MLP':
                ok = self.MLData_cb.MultiLayerPerceptron()
            elif combineAlgorithm == 'LR':
                ok = self.MLData_cb.LogisticRegressionClassifier()
            elif combineAlgorithm == 'LDA':
                ok = self.MLData_cb.LDAClassifier()
            elif combineAlgorithm == 'QDA':
                ok = self.MLData_cb.QDAClassifier()
            elif combineAlgorithm == 'KNN':
                ok = self.MLData_cb.KNeighbors()
            elif combineAlgorithm == 'LightGBM':
                ok = self.MLData_cb.LightGBMClassifier()
            elif combineAlgorithm == 'XGBoost':
                ok = self.MLData_cb.XGBoostClassifier()
            elif combineAlgorithm == 'SGD':
                ok = self.MLData_cb.StochasticGradientDescentClassifier()
            elif combineAlgorithm == 'DecisionTree':
                ok = self.MLData_cb.DecisionTree()
            elif combineAlgorithm == 'NaiveBayes':
                ok = self.MLData_cb.GaussianNBClassifier()
            elif combineAlgorithm == 'AdaBoost':
                ok = self.MLData_cb.AdaBoost()
            elif combineAlgorithm == 'Bagging':
                ok = self.MLData_cb.Bagging()
            elif combineAlgorithm == 'GBDT':
                ok = self.MLData_cb.GBDTClassifier()

            end_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')            
            if ok:
                self.status_label.setText('Training combined model for combination %s complete.' %str(c))
                msg = '%s\tTraining combined model for combination %s complete.' %(end_time, str(c))
                self.append_msg_signal.emit(msg)
                if self.metrics.classification_task == 'binary':
                    if self.MLData_cb.metrics.loc['Mean', 'AUROC'] > self.bestPerformance:
                        self.bestPerformance = self.MLData_cb.metrics.loc['Mean', 'AUROC']                           
                        self.bestMetrics = self.MLData_cb.metrics
                        self.bestAUC = self.MLData_cb.meanAucData
                        self.bestPRC = self.MLData_cb.meanPrcData
                        self.bestModels = self.MLData_cb.best_model
                        self.bestTrainingScore = self.MLData_cb.training_score
                        self.boxplot_data['Combined'] = self.MLData_cb.metrics.iloc[0:self.ml_defatult_para['FOLD'], :]

                else:
                    if self.MLData_cb.metrics.loc['Mean', 'AUROC'] > self.bestPerformance:
                        self.bestPerformance = self.MLData_cb.metrics.loc['Mean', 'Acc']
                        self.bestMetrics = self.MLData_cb.metrics
                        self.bestAUC = self.MLData_cb.meanAucData
                        self.bestPRC = self.MLData_cb.meanPrcData
                        self.bestModels = self.MLData_cb.best_model
                        self.bestTrainingScore = self.MLData_cb.training_score
                        self.boxplot_data['Combined'] = self.MLData_cb.metrics.iloc[0:self.ml_defatult_para['FOLD'], :]

        data_list = ['Combined_model', start_time, end_time] + list(self.bestMetrics.values[-1])
        item = pd.DataFrame([data_list], columns=['Id', 'StartTime', 'EndTime', 'Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC'])
        self.metrics.insert_data(item, 'Combined_model', self.bestAUC, self.bestPRC, self.bestTrainingScore, self.bestModels)

        msg = '%s\tThe combination with best performance is %s' % (end_time, str(c))
        self.append_msg_signal.emit(msg)

        self.display_signal.emit(data_list)
        self.display_curves_signal.emit()
        self.setDisabled(False)
        self.combineModelBtn.setDisabled(True)

    def append_message(self, message):
        self.logTextEdit.append(message)

    def display_correlation_heatmap(self):
        dataframe = self.metrics.metrics.iloc[:, 3:]
        dataframe.index = self.metrics.metrics.iloc[:, 0].values
        try:
            if self.MLData.task == 'binary':
                self.corrWindow = PlotWidgets.HeatmapWidget(dataframe)
                self.corrWindow.show()
            else:
                QMessageBox.critical(self, 'Error', 'Correlation can not be calculated for muti-classification task.', QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)
        except Exception as e:
            QMessageBox.critical(self, 'Error', str(e), QMessageBox.Ok | QMessageBox.No, QMessageBox.Ok)

    def closeEvent(self, event):
        reply = QMessageBox.question(self, 'Confirm Exit', 'Are you sure want to quit iLearnPlus?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.close_signal.emit('AutoML')
            self.close()
        else:
            if event:
                event.ignore()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setFont(QFont('Arial', 10))
    win = ILearnPlusAutoML()
    win.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
    win.show()
    sys.exit(app.exec_())


#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import numpy as np
import pandas as pd
import math
from sklearn.metrics import roc_curve, auc, precision_recall_curve

class Metrics(object):
    def __init__(self, scores, labels, pos_label=None, threshold=0.5):
        super(Metrics, self).__init__()
        self.scores = scores
        label_sets = sorted(set(labels))
        if pos_label is None:
            self.pos_label = label_sets[-1]
        else:
            self.pos_label = pos_label
        self.labels = labels
        self.threshold = threshold
        # evaluation metrics
        self.sensitivity = 0
        self.specificity = 0
        self.accuracy = 0
        self.mcc = 0
        self.precision = 0
        self.f1 = 0
        self.auc = 0
        self.prc = 0
        self.aucDot = None
        self.prcDot = None

        if len(label_sets) == 2:
            self.calculateBinaryTaskMetrics()
        elif len(label_sets) > 2:
            self.calculateMultiTaskMetrics()
        else:
            pass

    def calculateBinaryTaskMetrics(self):
        tp, tn, fp, fn = 0, 0, 0, 0
        for i in range(len(self.scores)):
            if self.labels[i] == self.pos_label:
                if self.scores[i] >= self.threshold:
                    tp += 1
                else:
                    fn += 1
            else:
                if self.scores[i] < self.threshold:
                    tn += 1
                else:
                    fp += 1

        self.sensitivity = round(tp / (tp + fn) * 100, 2) if (tp + fn) != 0 else 0
        self.specificity = round(tn / (fp + tn) * 100, 2) if (fp + tn) != 0 else 0
        self.accuracy = round((tp + tn) / (tp + fn + tn + fp) * 100, 2)
        self.mcc = round((tp * tn - fp * fn) / math.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)), 4) if (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn) != 0 else 0
        self.precision = round(tp / (tp + fp) * 100, 2) if (tp + fp) != 0 else 0
        self.f1 = round(2 * tp / (2 * tp + fp + fn), 4) if (2 * tp + fp + fn) != 0 else 0
        # roc        
        fpr, tpr, _ = roc_curve(self.labels, self.scores, pos_label=self.pos_label)
        self.auc = round(auc(fpr, tpr), 4)
        self.aucDot = pd.DataFrame(np.hstack((fpr.reshape((-1, 1)), tpr.reshape((-1, 1)))), columns=['fpr', 'tpr'])
        # prc
        precision, recall, _ = precision_recall_curve(self.labels, self.scores, pos_label=self.pos_label)
        self.prc = round(auc(recall, precision), 4)
        self.prcDot = pd.DataFrame(np.hstack((recall.reshape((-1, 1)), precision.reshape((-1, 1)))),
                                   columns=['recall', 'precision'])

    def calculateMultiTaskMetrics(self):
        result = []
        for item in self.scores:
            result.append(max(enumerate(item), key=lambda x: x[1])[0])
        result = np.array(result)
        self.accuracy = round(len(result[result == self.labels]) / len(self.labels) * 100, 2)

    @staticmethod
    def getBinaryTaskMetrics(scores, labels, pos_label=None, threshold=0.5):
        evaluationMetrics = Metrics(scores, labels, pos_label, threshold)
        return (evaluationMetrics.sensitivity, evaluationMetrics.specificity, evaluationMetrics.precision, evaluationMetrics.accuracy, evaluationMetrics.mcc,
                evaluationMetrics.f1, evaluationMetrics.auc, evaluationMetrics.prc), True

    @staticmethod
    def getMutiTaskMetrics(scores, labels):
        evaluationMetricss = Metrics(scores, labels)
        return evaluationMetricss.accuracy, True


if __name__ == '__main__':
    data = np.loadtxt('SVM_IND.txt', delimiter='\t')
    metrics = Metrics(data[:, 1:], data[:, 0])
    print(metrics.aucDot)

    # acc = Metrics.getMutiTaskMetrics(data[:, 1:], data[:, 0])
    # print(acc)



#!/usr/bin/env python
# _*_ coding: utf-8 _*_

import os, sys, re, platform

pPath = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(pPath)

import warnings
import pandas as pd
import numpy as np
import torch
from sklearn.naive_bayes import GaussianNB
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, BaggingClassifier, GradientBoostingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn import svm
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.model_selection import StratifiedKFold, GridSearchCV
import lightgbm as lgb
import xgboost as xgb
from scipy import interp
from EvaluationMetrics import Metrics
from Nets import (DealDataset, Net_CNN_1, Net_CNN_11, Net_RNN_2, Net_ABCNN_4, Net_ResNet_5, Net_AutoEncoder_6)
from torch.utils.data import DataLoader
warnings.filterwarnings("ignore", category=FutureWarning, module="sklearn")
warnings.filterwarnings("ignore", category=DeprecationWarning, module="sklearn")
warnings.filterwarnings("ignore", category=UserWarning, module="sklearn")


class ILearnMachineLearning(object):
    def __init__(self, kw):
        self.kw = kw                                     # parameters dict
        self.training_dataframe = None                   # Pandas DataFrame
        self.training_datalabel = None                   # ndarray
        self.testing_dataframe = None                    # Pandas DataFrame
        self.testing_datalabel = None                    # ndarray
        self.training_score = None                       # 2D ndarray [fold, label, scores...]
        self.testing_score = None                        # 2D ndarray [0, label, scores...]
        self.best_model = None                           # list
        self.best_n_trees = 0                            # int
        self.metrics = None                              # Pandas DataFrame
        self.aucData = None                              # list
        self.prcData = None                              # list
        self.meanAucData = None                          # Pandas DataFrame
        self.meanPrcData = None                          # Pandas DataFrame
        self.indepAucData = None                         # Pandas DataFrame
        self.indepPrcData = None                         # Pandas DataFrame
        self.algorithm = None                            # string
        self.error_msg = None                            # string
        self.boxplot_data = None                         # dict key -> metrics.columns  value -> BoxPlotData object
        self.message = None                              # string
        self.task = None                                 # string 'binary' or 'muti-task'

    def load_data(self, file, target='Training'):
        if not os.path.exists(file):
            self.error_msg = 'data file does not exist.'
            return False
        dataframe, datalabel = None, None
        try:
            if file.endswith('.tsv'):
                df = pd.read_csv(file, sep='\t', header=None)
                dataframe = df.iloc[:, 1:]
                dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                datalabel = np.array(df.iloc[:, 0]).astype(int)
            elif file.endswith('.csv'):
                df = pd.read_csv(file, sep=',', header=None)
                dataframe = df.iloc[:, 1:]
                dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                datalabel = np.array(df.iloc[:, 0]).astype(int)
            elif file.endswith('.svm'):
                with open(file) as f:
                    record = f.read().strip()
                record = re.sub('\d+:', '', record)
                array = np.array([[i for i in item.split()] for item in record.split('\n')])
                dataframe = pd.DataFrame(array[:, 1:], dtype=float)
                dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                datalabel = array[:, 0].astype(int)
            else:
                with open(file) as f:
                    record = f.read().strip().split('@')[-1].split('\n')[1:]
                array = np.array([item.split(',') for item in record])
                dataframe = pd.DataFrame(array[:, 0:-1], dtype=float)
                dataframe.index=['Sample_%s'%i for i in range(dataframe.values.shape[0])]
                dataframe.columns = ['F_%s'%i for i in range(dataframe.values.shape[1])]
                label = []
                for i in array[:, -1]:
                    if i == 'yes':
                        label.append(1)
                    else:
                        label.append(0)
                datalabel = np.array(label)
            if target == 'Training':
                self.training_dataframe = dataframe
                self.training_datalabel = datalabel
                if len(set(datalabel)) == 2:
                    self.task = 'binary'
                if len(set(datalabel)) > 2:
                    self.task = 'muti-task'
            if target == 'Testing':
                self.testing_dataframe = dataframe
                self.testing_datalabel = datalabel
        except Exception as e:
            self.error_msg = str(e)
            return False
        return True

    def import_training_data(self, dataframe, label):
        self.training_dataframe = dataframe
        self.training_datalabel = label
        if len(set(label)) == 2:
            self.task = 'binary'
        if len(set(label)) > 2:
            self.task = 'muti-task'

    def import_testing_data(self, dataframe, label):
        self.testing_dataframe = dataframe
        self.testing_datalabel = label

    def RandomForest(self):
        try:
            self.algorithm = 'RF'
            auto = self.kw['auto']
            fold = self.kw['FOLD']
            n_trees = self.kw['n_trees']
            n_jobs = self.kw['cpu']
            if auto:
                tree_range = self.kw['tree_range']
            else:
                tree_range = (n_trees, n_trees + 1, 100)
        
            if not self.training_dataframe is None:
                categories = sorted(set(self.training_datalabel))
                X, y = self.training_dataframe.values, self.training_datalabel
                # best model selection
                best_n_trees = tree_range[0]
                best_auc = 0
                best_accuracy = 0
                best_model = []
                best_training_score = None
                for tree in range(tree_range[0], tree_range[1] + 1, tree_range[2]):
                    training_score = np.zeros((X.shape[0], len(categories) + 2))
                    training_score[:, 1] = y
                    model = []
                    folds = StratifiedKFold(fold).split(X, y) # python iterator can be used only once
                    for i, (train, valid) in enumerate(folds):
                        train_X, train_y = X[train], y[train]
                        valid_X, valid_y = X[valid], y[valid]
                        rfc_model = RandomForestClassifier(n_estimators=tree, bootstrap=False, n_jobs=n_jobs)
                        rfc = rfc_model.fit(train_X, train_y)
                        model.append(rfc)
                        training_score[valid, 0] = i
                        training_score[valid, 2:] = rfc.predict_proba(valid_X)

                    if len(categories) == 2:
                        metrics, ok = Metrics.getBinaryTaskMetrics(training_score[:, 3], training_score[:, 1], pos_label=categories[-1])
                        if metrics[6] > best_auc:
                            best_auc = metrics[6]
                            best_n_trees = tree
                            best_model = model
                            best_training_score = training_score
                    if len(categories) > 2:
                        metrics = Metrics.getMutiTaskMetrics(training_score[:, 2:], training_score[:, 1])
                        if metrics[0] > best_accuracy:
                            best_accuracy = metrics[0]
                            best_n_trees = tree
                            best_model = model
                            best_training_score = training_score

                self.training_score = pd.DataFrame(best_training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.best_model = best_model
                self.best_n_trees = best_n_trees

                # independent dataset
                if not self.testing_dataframe is None:
                    indep = self.testing_dataframe.values
                    testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                    testing_score[:, 1] = self.testing_datalabel
                    for rfc in self.best_model:
                        testing_score[:, 2:] += rfc.predict_proba(indep)
                    testing_score[:, 2:] /= fold
                    self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                    self.testing_score['Fold'] = 'NA'
                # calculate metrics
                self.calculateEvaluationMetrics()
                self.message = 'Best n_trees is %d' %self.best_n_trees
                return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def SupportVectorMachine(self):
        try:
            self.algorithm = 'SVM'
            kernel = self.kw['kernel']
            auto = self.kw['auto']
            fold= self.kw['FOLD']
            penality = self.kw['penality']
            gamma = 1/self.training_dataframe.values.shape[1] if self.kw['gamma'] == 'auto' else self.kw['gamma']
            penalityRange = self.kw['penalityRange']
            gammaRange = self.kw['gammaRange']

            parameters = {'kernel': ['linear'], 'C': penalityRange} if kernel == 'linear' else {'kernel': [kernel], 'C': penalityRange, 'gamma': 2.0 ** np.arange(gammaRange[0], gammaRange[1])}
        
            if auto:
                optimizer = GridSearchCV(svm.SVC(probability=True), parameters).fit(self.training_dataframe.values, self.training_datalabel)
                params = optimizer.best_params_
                penality = params['C']
                if kernel != 'linear':
                    gamma = params['gamma']
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                svm_model = svm.SVC(C=penality, kernel=kernel, degree=3, gamma=gamma, coef0=0.0, shrinking=True, probability=True, random_state=1)
                svc = svm_model.fit(train_X, train_y)
                model.append(svc)
                training_score[valid, 0] = i
                training_score[valid, 2:] = svc.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for svc in self.best_model:
                    testing_score[:, 2:] += svc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'penality: %s' %penality
            self.message += ' gamma: %s' %gamma if kernel != 'linear' else ''
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def MultiLayerPerceptron(self):
        try:
            self.algorithm = 'MLP'
            hidden_layer_size = tuple([int(i) for i in self.kw['layer'].split(';')])
            for item in hidden_layer_size:
                if item > 256:
                    self.error_msg = 'The layer size is out of range (1~256).'
                    return False

            activation = self.kw['activation']
            optimizer = self.kw['optimizer']
            epochs = self.kw['epochs']
            fold = self.kw['FOLD']
            lr=0.001
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                mlp_model = MLPClassifier(activation=activation, alpha=1e-05, batch_size='auto', beta_1=0.9, beta_2=0.999,
                                      early_stopping=False, epsilon=1e-08, hidden_layer_sizes=hidden_layer_size,
                                      learning_rate='constant', learning_rate_init=lr, max_iter=epochs, momentum=0.9,
                                      nesterovs_momentum=True, power_t=0.5, random_state=1, shuffle=True, solver=optimizer,
                                      tol=0.0001, validation_fraction=0.1, verbose=False, warm_start=False)
                mlp = mlp_model.fit(train_X, train_y)
                model.append(mlp)
                training_score[valid, 0] = i
                training_score[valid, 2:] = mlp.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for mlp in self.best_model:
                    testing_score[:, 2:] += mlp.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def LogisticRegressionClassifier(self):
        try:
            self.algorithm = 'LR'
            fold = self.kw['FOLD']        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                lr_model = LogisticRegression(C=1.0, random_state=0).fit(train_X, train_y)
                model.append(lr_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = lr_model.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for lrc in self.best_model:
                    testing_score[:, 2:] += lrc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def LDAClassifier(self):
        try:
            self.algorithm = 'LDA'
            fold = self.kw['FOLD']        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                lda_model = LinearDiscriminantAnalysis(solver='svd', store_covariance=True).fit(train_X, train_y)
                model.append(lda_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = lda_model.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score,
                                               columns=['Fold', 'Label'] + ['Score_%s' % i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for lda in self.best_model:
                    testing_score[:, 2:] += lda.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score,
                                                  columns=['Fold', 'Label'] + ['Score_%s' % i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def QDAClassifier(self):
        try:
            self.algorithm = 'QDA'
            fold = self.kw['FOLD']        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                qda_model = QuadraticDiscriminantAnalysis(store_covariance=True).fit(train_X, train_y)
                model.append(qda_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = qda_model.predict_proba(valid_X)            
            self.training_score = pd.DataFrame(training_score,
                                                columns=['Fold', 'Label'] + ['Score_%s' % i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for qda in self.best_model:
                    testing_score[:, 2:] += qda.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score,
                                                    columns=['Fold', 'Label'] + ['Score_%s' % i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def KNeighbors(self):
        try:
            self.algorithm = 'KNN'
            fold = self.kw['FOLD']
            n_neighbors = self.kw['topKValue']        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                knn_model = KNeighborsClassifier(n_neighbors=n_neighbors).fit(train_X, train_y)
                model.append(knn_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = knn_model.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for lrc in self.best_model:
                    testing_score[:, 2:] += lrc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def LightGBMClassifier(self):
        try:
            self.algorithm = 'LightGBM'
            fold= self.kw['FOLD']
            boosting_type = self.kw['boosting_type']
            num_leaves = self.kw['num_leaves']
            max_depth = self.kw['max_depth']
            learning_rate = self.kw['learning_rate']
            leaves_range = self.kw['leaves_range']
            depth_range = self.kw['depth_range']
            rate_range = self.kw['rate_range']
            auto = self.kw['auto']
            n_jobs = self.kw['cpu']
            
            parameters = {
                'num_leaves': list(range(leaves_range[0], leaves_range[1], leaves_range[2])),
                'max_depth': list(range(depth_range[0], depth_range[1], depth_range[2])),
                'learning_rate': list(np.arange(rate_range[0], rate_range[1], rate_range[2]))
            }
            if auto:
                gbm = lgb.LGBMClassifier(boosting_type=boosting_type)
                gsearch = GridSearchCV(gbm, param_grid=parameters, n_jobs=n_jobs).fit(self.training_dataframe.values, self.training_datalabel)
                best_parameters = gsearch.best_params_
                num_leaves = best_parameters['num_leaves']
                max_depth = best_parameters['max_depth']
                learning_rate = best_parameters['learning_rate']

            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                gbm_model = lgb.LGBMClassifier(boosting_type=boosting_type, num_leaves=num_leaves, max_depth=max_depth, learning_rate=learning_rate).fit(train_X, train_y)
                model.append(gbm_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = gbm_model.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for gbm in self.best_model:
                    testing_score[:, 2:] += gbm.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            if auto:
                self.message = 'Best parameters: ' + str(best_parameters)
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def XGBoostClassifier(self):
        try:
            self.algorithm = 'XGBoost'
            fold= self.kw['FOLD']
            booster = self.kw['booster']
            max_depth = self.kw['max_depth']
            learning_rate = self.kw['learning_rate']
            depth_range = self.kw['depth_range']
            rate_range = self.kw['rate_range']
            auto = self.kw['auto']
            n_jobs = self.kw['cpu']
            n_estimator = self.kw['n_estimator']
            colsample_bytree = self.kw['colsample_bytree']
            parameters = {
                'max_depth': list(range(depth_range[0], depth_range[1], depth_range[2])),
                'learning_rate': list(np.arange(rate_range[0], rate_range[1], rate_range[2]))
            }
        
            if auto:
                bst = xgb.XGBClassifier(booster=booster, n_estimators=n_estimator, n_jobs=n_jobs, colsample_bytree=colsample_bytree)
                gsearch = GridSearchCV(bst, param_grid=parameters, n_jobs=n_jobs).fit(self.training_dataframe.values, self.training_datalabel)
                best_parameters = gsearch.best_params_
                max_depth = best_parameters['max_depth']
                learning_rate = best_parameters['learning_rate']
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                bst_model = xgb.XGBClassifier(booster=booster, max_depth=max_depth, learning_rate=learning_rate, n_estimators=n_estimator, n_jobs=n_jobs, colsample_bytree=colsample_bytree).fit(train_X, train_y)
                model.append(bst_model)
                training_score[valid, 0] = i
                training_score[valid, 2:] = bst_model.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for bst in self.best_model:
                    testing_score[:, 2:] += bst.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            if auto:
                self.message = 'Best parameters: ' + str(best_parameters)
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def StochasticGradientDescentClassifier(self):
        try:
            self.algorithm = 'SGD'
            fold = self.kw['FOLD']
        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                sgd = SGDClassifier(loss='log').fit(train_X, train_y)
                model.append(sgd)
                training_score[valid, 0] = i
                training_score[valid, 2:] = sgd.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for sgd in self.best_model:
                    testing_score[:, 2:] += sgd.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def DecisionTree(self):
        try:
            self.algorithm = 'DecisionTree'
            fold = self.kw['FOLD']
        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                dtc = DecisionTreeClassifier().fit(train_X, train_y)
                model.append(dtc)
                training_score[valid, 0] = i
                training_score[valid, 2:] = dtc.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for dtc in self.best_model:
                    testing_score[:, 2:] += dtc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def GaussianNBClassifier(self):
        try:
            self.algorithm = 'Bayes'
            fold = self.kw['FOLD']
        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                gsn = GaussianNB().fit(train_X, train_y)
                model.append(gsn)
                training_score[valid, 0] = i
                training_score[valid, 2:] = gsn.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for gsn in self.best_model:
                    testing_score[:, 2:] += gsn.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def AdaBoost(self):
        try:
            self.algorithm = 'AdaBoost'
            fold = self.kw['FOLD']
        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                abc = AdaBoostClassifier(n_estimators=200, random_state=0).fit(train_X, train_y)
                model.append(abc)
                training_score[valid, 0] = i
                training_score[valid, 2:] = abc.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for abc in self.best_model:
                    testing_score[:, 2:] += abc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def Bagging(self):
        try:
            self.algorithm = 'Bagging'
            fold = self.kw['FOLD']
            n_estimators = self.kw['n_estimator']
            n_jobs = self.kw['cpu']


            
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                bgc = BaggingClassifier(n_estimators=n_estimators, n_jobs=n_jobs).fit(train_X, train_y)
                model.append(bgc)
                training_score[valid, 0] = i
                training_score[valid, 2:] = bgc.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for bgc in self.best_model:
                    testing_score[:, 2:] += bgc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def GBDTClassifier(self):
        try:
            self.algorithm = 'GBDT'
            fold = self.kw['FOLD']
        
            categories = sorted(set(self.training_datalabel))
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], len(categories) + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_X, train_y = X[train], y[train]
                valid_X, valid_y = X[valid], y[valid]
                gbc = GradientBoostingClassifier().fit(train_X, train_y)
                model.append(gbc)
                training_score[valid, 0] = i
                training_score[valid, 2:] = gbc.predict_proba(valid_X)
            self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
            self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], len(categories) + 2))
                testing_score[:, 1] = self.testing_datalabel
                for gbc in self.best_model:
                    testing_score[:, 2:] += gbc.predict_proba(indep)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'
            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def calculateEvaluationMetrics(self):
        if not self.training_score is None:
            fold = sorted(set(self.training_score['Fold']))
            data = self.training_score.values
            categories = sorted(set(self.training_datalabel))
            index_name = []
            column_name = ['Sn', 'Sp', 'Pre', 'Acc', 'MCC', 'F1', 'AUROC', 'AUPRC']
            if len(categories) == 2:
                dataMetrics = np.zeros((len(fold), 8))
                aucData = []
                prcData = []
                # data for plot mean ROC in cross-validation
                mean_fpr = np.linspace(0, 1, 100)
                tprs = []
                # data for plot mean PRC in cross-validation
                mean_recall = np.linspace(0, 1, 100)
                precisions = []

                for f in fold:
                    tmp_data = data[data[:, 0] == f]
                    metrics = Metrics(tmp_data[:, -1], tmp_data[:, 1])
                    aucData.append(['Fold %d AUROC = %s' %(f, metrics.auc), metrics.aucDot])
                    prcData.append(['Fold %d AUPRC = %s' %(f, metrics.prc), metrics.prcDot])
                    tprs.append(interp(mean_fpr, metrics.aucDot['fpr'], metrics.aucDot['tpr']))
                    tprs[-1][0] = 0
                    precisions.append(interp(mean_recall, metrics.prcDot['recall'][::-1], metrics.prcDot['precision'][::-1]))
                    dataMetrics[int(f)] = np.array([metrics.sensitivity, metrics.specificity, metrics.precision, metrics.accuracy, metrics.mcc, metrics.f1, metrics.auc, metrics.prc]).reshape((1, -1))
                    index_name.append('Fold %d' %f)
                    del metrics
                meanValue = np.around(np.mean(dataMetrics, axis=0), decimals=4)
                dataMetrics = np.vstack((dataMetrics, meanValue))
                index_name.append('Mean')
                self.aucData = aucData
                self.prcData = prcData
                mean_tpr = np.mean(tprs, axis=0)
                mean_tpr[-1] = 1.0
                mean_precision = np.mean(precisions, axis=0)
                self.meanAucData = ['Mean AUROC = %s' %dataMetrics[-1][-2], pd.DataFrame(np.hstack((mean_fpr.reshape((-1, 1)), mean_tpr.reshape((-1, 1)))), columns=['fpr', 'tpr'])]
                self.meanPrcData = ['Mean AUPRC = %s' %dataMetrics[-1][-1], pd.DataFrame(np.hstack((mean_recall.reshape((-1, 1)), mean_precision.reshape((-1, 1)))), columns=['recall', 'precision'])]

                if not self.testing_score is None:
                    data = self.testing_score.values[:, 1:].astype(float)
                    metrics = Metrics(data[:, -1], data[:, 0])
                    metrics_ind = np.array([metrics.sensitivity, metrics.specificity, metrics.precision, metrics.accuracy, metrics.mcc, metrics.f1, metrics.auc, metrics.prc]).reshape((1, -1))
                    dataMetrics = np.vstack((dataMetrics, metrics_ind))
                    index_name.append('Indep')
                    self.indepAucData = ['Indep AUROC = %s' %metrics.auc, metrics.aucDot]
                    self.indepPrcData = ['Indep AUPRC = %s' %metrics.prc, metrics.prcDot]
                    del metrics
                self.metrics = pd.DataFrame(dataMetrics, index=index_name, columns=column_name)
            elif len(categories) > 2:
                dataMetrics = np.zeros((len(fold), 8)).astype(str)
                for f in fold:
                    index_name.append('Fold %d' %f)
                    tmp_data = data[data[:, 0] == f]
                    metrics = Metrics(tmp_data[:, 2:], tmp_data[:, 1])
                    dataMetrics[int(f)] = np.array(['NA', 'NA', 'NA', '%s' %metrics.accuracy, 'NA', 'NA', 'NA', 'NA'])
                    del metrics
                self.metrics = pd.DataFrame(dataMetrics, index=index_name, columns=column_name)
            else:
                pass

    def save_prediction_score(self, file, type='training'):
        try:
            if type == 'training':
                df = pd.DataFrame(self.training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in sorted(set(self.training_datalabel))])
                df.to_csv(file, sep='\t', header=True, index=False)
            else:
                df = pd.DataFrame(self.testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in sorted(set(self.training_datalabel))])
                df.to_csv(file, sep='\t', header=True, index=False)
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def save_metrics(self, file):
        try:
            self.metrics.to_csv(file, sep='\t', header=True, index=True)
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def save_coder(self, file, type='training'):
        try:
            if type == 'training':
                data = np.hstack((self.training_datalabel.reshape((-1, 1)), self.training_dataframe.values))
            else:
                data = np.hstack((self.testing_datalabel.reshape((-1, 1)), self.testing_dataframe.values))
            if file.endswith('.csv'):
                np.savetxt(file, data, fmt="%s", delimiter=',')
            if file.endswith('.tsv'):
                np.savetxt(file, data, fmt="%s", delimiter=',')
            if file.endswith('.svm'):
                with open(file, 'w') as f:
                    for line in data:
                        f.write('%s' % line[0])
                        for i in range(1, len(line)):
                            f.write('  %d:%s' % (i, line[i]))
                        f.write('\n')
            if file.endswith('.arff'):
                with open(file, 'w') as f:
                    f.write('@relation descriptor\n\n')
                    for i in range(1, len(data[0])):
                        f.write('@attribute f.%d numeric\n' % i)
                    f.write('@attribute play {yes, no}\n\n')
                    f.write('@data\n')
                    for line in data:
                        for fea in line[1:]:
                            f.write('%s,' % fea)
                        if int(line[0]) == 1:
                            f.write('yes\n')
                        else:
                            f.write('no\n')
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False

    def calculate_boxplot_data(self):
        df = self.metrics.iloc[0:self.kw['FOLD']]
        task = len(set(self.training_datalabel))

        boxplot_dict = {}
        if task == 2:
            for item in df.columns:
                boxplot_dict[item] = BoxPlotData(df.loc[:, item])
        if task >= 3:
            boxplot_dict['Acc'] = BoxPlotData(df.loc[:, 'Acc'].values.astype(float))
        self.boxplot_data = boxplot_dict

    """ run deep learning frames """
    def to_categorical(self, y, num_classes=None):
        y = np.array(y, dtype='int')
        input_shape = y.shape
        if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
            input_shape = tuple(input_shape[:-1])
        y = y.ravel()
        if not num_classes:
            num_classes = np.max(y) + 1
        n = y.shape[0]
        categorical = np.zeros((n, num_classes))
        categorical[np.arange(n), y] = 1
        output_shape = input_shape + (num_classes,)
        categorical = np.reshape(categorical, output_shape)
        return categorical

    def run_networks(self, network=1):
        try:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            fold = self.kw['FOLD']
            batch_size = self.kw['batch_size']
            categories = sorted(set(self.training_datalabel))
            categories_onehot = self.to_categorical(categories)         
            
            X, y = self.training_dataframe.values, self.training_datalabel
            training_score = np.zeros((X.shape[0], categories_onehot.shape[1] + 2))
            training_score[:, 1] = y
            model = []
            folds = StratifiedKFold(fold).split(X, y)
            for i, (train, valid) in enumerate(folds):
                train_set = DealDataset(X[train], y[train])
                valid_set = DealDataset(X[valid], y[valid])
                train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
                valid_loader = DataLoader(valid_set, batch_size=batch_size, shuffle=False)
                # if network == 1 and len(categories) > 2:
                #     self.algorithm = 'CNN'
                #     net = Net_CNN_1(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], out_channel=self.kw['output_channel'], padding=self.kw['padding'], conv_kernel_size=self.kw['kernel_size'], dense_size=self.kw['fc_size'], dropout=self.kw['dropout']).to(device)
                #     net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                # elif network == 1 and len(categories) == 2:
                #     self.algorithm = 'CNN'                
                #     net = Net_CNN_11(device=device, input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], out_channel=self.kw['output_channel'], padding=self.kw['padding'], conv_kernel_size=self.kw['kernel_size'], dense_size=self.kw['fc_size'], dropout=self.kw['dropout']).to(device)
                #     net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                if network == 1:
                    self.algorithm = 'CNN'
                    net = Net_CNN_1(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], out_channel=self.kw['output_channel'], padding=self.kw['padding'], conv_kernel_size=self.kw['kernel_size'], dense_size=self.kw['fc_size'], dropout=self.kw['dropout']).to(device)
                    net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                elif network == 2:
                    self.algorithm = 'RNN'
                    net = Net_RNN_2(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], hidden_size=self.kw['rnn_hidden_size'], num_layers=self.kw['rnn_hidden_layers'], dense_size=self.kw['fc_size'], dropout=self.kw['dropout']).to(device)
                    net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                elif network == 3:
                    self.algorithm = 'BCNN'
                    net = Net_RNN_2(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], hidden_size=self.kw['rnn_hidden_size'], num_layers=self.kw['rnn_hidden_layers'], dense_size=self.kw['fc_size'], dropout=self.kw['dropout'], bidirectional=True).to(device)
                    net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                elif network == 4:
                    self.algorithm = 'ABCNN'
                    net = Net_ABCNN_4(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length'], dropout=self.kw['dropout']).to(device)
                    net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                elif network == 5:
                    self.algorithm = 'ResNet'
                    net = Net_ResNet_5(device=device, category=categories_onehot.shape[1], input_size=self.kw['input_channel'], sequence_length=self.kw['input_length']).to(device)
                    net.fit(train_loader, valid_loader)
                elif network == 6:
                    self.algorithm = 'AE'
                    net = Net_AutoEncoder_6(device=device, category=categories_onehot.shape[1], input_dim=self.training_dataframe.values.shape[1]).to(device)
                    net.fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
                    net.re_build_net()
                    net.re_fit(train_loader, valid_loader, epochs=self.kw['epochs'], early_stopping=self.kw['early_stopping'], lr=self.kw['learning_rate'])
            
                model.append(net)
                training_score[valid, 0] = i
                training_score[valid, 2:] = net.predict(valid_loader)
                self.training_score = pd.DataFrame(training_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.best_model = model

            # independent dataset
            if not self.testing_dataframe is None:
                indep = self.testing_dataframe.values
                testing_score = np.zeros((indep.shape[0], categories_onehot.shape[1] + 2))
                testing_score[:, 1] = self.testing_datalabel
                indep_set = DealDataset(indep, self.testing_datalabel)
                indep_loader = DataLoader(indep_set, batch_size, shuffle=False)
                for net in self.best_model:
                    testing_score[:, 2:] += net.predict(indep_loader)
                testing_score[:, 2:] /= fold
                self.testing_score = pd.DataFrame(testing_score, columns=['Fold', 'Label'] + ['Score_%s' %i for i in categories])
                self.testing_score['Fold'] = 'NA'

            # calculate metrics
            self.calculateEvaluationMetrics()
            self.message = 'Complete.'
            return True
        except Exception as e:
            self.error_msg = str(e)
            return False


class BoxPlotData(object):
    def __init__(self, data, whis=1.5):
        super(BoxPlotData, self).__init__()
        self.data = data
        self.whis = whis
        self.boxplot_dict = self.calculate()
    def calculate(self):
        try:
            median_value = np.median(self.data)
            Q1 = None
            Q3 = None
            i = len(self.data) // 4
            a = sorted(self.data)
            if len(self.data) % 4 == 0:
                Q1 = a[i-1] * 0.25 + a[i] * 0.75
                Q3 = a[3*i-1] * 0.75 + a[3*i] * 0.25
            elif len(self.data) % 4 == 1:
                Q1 = a[i]
                Q3 = a[3*i]
            elif len(self.data) % 4 == 2:
                Q1 = a[i] * 0.75 + a[i+1] * 0.25
                Q3 = a[3*i] * 0.25 + a[3*i+1] * 0.75
            elif len(self.data) % 4 == 3:
                Q1 = a[i] * 0.5 + a[i+1] * 0.5
                Q3 = a[3*i+1] * 0.5 + a[3*i+2] * 0.5
            IQR = (Q3 - Q1) * self.whis
            outlier = []
            Max_value = Q3 + IQR
            Min_value = Q1 - IQR
            for item in a:
                if item >= (Q1 - IQR):
                    Min_value = item
                    break
                else:
                    outlier.append(item)
            for item in a[::-1]:
                if item <= (Q3 + IQR):
                    Max_value = item
                    break
                else:
                    outlier.append(item)
            return {'median': median_value, 'Q1': Q1, 'Q3': Q3, 'maximum': Max_value, 'minimum': Min_value, 'outlier': outlier}
        except Exception as e:
            return {}


if __name__ == '__main__':
    ml_defatult_para = {
        'FOLD': 5,
        'cpu': 1,
        'auto': False,
        'n_trees': 100,
        'tree_range': (100, 1000, 100),
        'kernel': 'rbf',
        'penality': 1.0,
        'gamma': 'auto',
        'penalityRange': (1.0, 15.0),
        'gammaRange': (-10.0, 5.0),
        'layer': '32;32',
        'activation': 'relu',
        'optimizer': 'adam',
        'topKValue': 3,
        'boosting_type': 'gbdt',
        'num_leaves': 31,
        'max_depth': -1,
        'learning_rate': 0.01,
        'leaves_range': (20, 100, 10),
        'depth_range': (15, 55, 10),
        'rate_range': (0.01, 0.15, 0.02),
        'booster': 'gbtree',
        'n_estimator': 100,
        'colsample_bytree': 0.8,
        'input_channel': 4,
        'input_length': 101,
        'output_channel': 64,
        'padding': 2,
        'kernel_size': 5,
        'dropout': 0.5,
        'epochs': 1000,
        'early_stopping': 100,
        'batch_size': 64,
        'rnn_hidden_size': 32,
        'rnn_num_layers': 1,
        'fc_size': 64,
    }
    Data = ILearnMachineLearning(ml_defatult_para)
    Data.load_data('../data/binary.csv')
    Data.load_data('../data/binary_ind.csv', target='Testing')
    Data.run_networks(5)
    print(Data.metrics)










#!/usr/bin/env python
# -*- coding:utf-8 -*-
# Author: wxnacy(chenzhen-win2009@163.com)

from .iLearnPlusBasic import *
from .iLearnPlusEstimator import *
from .iLearnPlusAutoML import *
from .iLearnPlusLoadModel import *
from .util import *
from .iLearnPlus import runiLearnPlus




#!/usr/bin/env python
#_*_coding:utf-8_*_

__all__ = [
    'CheckAccPseParameter',
    'DataAnalysis',
    'EvaluationMetrics',
    'FileProcessing',
    'InputDialog',
    'MachineLearning',
    'MCL',
    'ModelMetrics',
    'Modules',
    'Nets',
    'PlotWidgets',
    'TableWidget',   
]
